var path = require('path');
var fs = require('fs');
var os = require('os');
var util = require('util');
var xml2js = require('xml2js');
var _ = require('underscore');
var csv = require('ya-csv');
var exec = require('child_process').exec;
var spawn = require('child_process').spawn;

var rootPath = "./";
var dotPath = "dot";

// onload
var defs = "";

var distribution = os.type();
function updatedefs(onclose){
    console.log("updating defs");
    var child;
    var env = process.env.FLOWDESIGNER_PATH || "";
    console.log("FLOWDESIGNER_PATH", env);

    // distribution decision
    if(distribution.indexOf("Linux") !== -1)   distribution = "Linux";
    if(distribution.indexOf("Windows") !== -1) distribution = "Windows";
    if(distribution.indexOf("Darwin") !== -1)  distribution = "Darwin";
    console.log("OS: " + distribution);

    if(distribution === "Linux" || distribution === "Darwin"){
        paths = env.split(":");
        paths.push("`flowdesigner-config --toolboxdir`");
        child = exec("find " + paths.join(" ") + " -name '*.def' ");
        child.stdout.on("data", function(data){ defs += data; });
        child.on("close", function(code, signal){
            defs = _.uniq(defs.split("\n"));
            defs = _.uniq(defs, function(def){return def.split(path.sep).slice(-1)[0];});
            defs = _.sortBy(defs, function(def){return def.split(path.sep).slice(-1)[0];});
            console.log(util.inspect(defs));
            if(_.isFunction(onclose)) onclose();});
    }
    else if(distribution === "Windows"){
        distribution = "Windows";
        console.log("OS: Windows");
        paths = env.split(";");
        cmd = 'dir /s /b ';
        _.each(paths, function(p){
            cmd += '"' + p + '"';
        });
        cmd += '| findstr -i \\.def$';

        console.log(cmd);
        child = exec(cmd);
        child.stdout.on("data", function(data){ defs += data; });
        child.on("close", function(code, signal){
            defs = _.uniq(defs.split("\r\n"));
            defs = _.uniq(defs, function(def){return def.split(path.sep).slice(-1)[0];});
            defs = _.sortBy(defs, function(def){return def.split(path.sep).slice(-1)[0];});
            console.log("found defs:", defs);
            if(_.isFunction(onclose)) onclose();
        });

        // find dot (graphviz binary). seems ugly. sync version of exec is needed.
        exec("dot", function(err, stdout, stderr){
            if(err){
                var child1 = exec('dir /s /b "C:\\Program Files" | findstr -i dot.exe | findstr -i graphviz', function(err, stdout, stderr){
                    if(err){
                        var child2 = exec('dir /s /b "C:\\Program Files (x86)" | findstr -i dot.exe | findstr -i graphviz', function(err, stdout, stderr){
                            if(err){
                                dotPath = false;
                            } else{
                                dotPath = '"' + stdout.slice(0, -2) + '"';
                                console.log("path to graphviz executable:", dotPath);
                            }
                        });
                    }
                    else{
                        dotPath = '"' + stdout.slice(0, -2) + '"';
                        console.log("path to graphviz executable:", dotPath);
                    }
                });
            }else{
                dotPath = "dot";
                console.log("path to graphviz executable:", dotPath);
            }
        });
    }
}

updatedefs();

exports.index = function (req, res) {
    console.log("root:", req.query.root);
    if(req.query.root === "hark-designer")
        rootPath = "hark-designer/";

    selenium = "right";
    if(req.query.selenium === "selenium")
        selenium = "left";
    // res.render('index', { title: 'Express' });
    // res.sendfile('public/index.html');
    var css = ['jquery/jquery-ui-1.10.2.custom/css/smoothness/jquery-ui-1.10.2.custom.min.css',
               'jquery/jquery.contextMenu.css',
               'js/intro.js-0.4.0/minified/introjs.min.css',
               'stylesheets/main.css'];
    var js =  [
               'js/json2.js',
               'jquery/jquery.min.1.9.1.js',
               'jquery/jquery-ui-1.10.2.custom/js/jquery-ui-1.10.2.custom.min.js',
               'jquery/jquery.ui.position.js',
               'jquery/jquery.contextMenu.js',
               'js/underscore.js',
               'js/arrayObjectMixin.js',
               'js/backbone.js',
               'js/fileBrowse.js',
               'js/util.js',
               'js/main.js',
               'js/intro.js-0.4.0/minified/intro.min.js'];

    css = _.map(css, function(tmp){return rootPath + tmp;});
    js  = _.map(js , function(tmp){return rootPath + tmp;});
    res.render('index', {css: css, javascripts: js,
        rootPath: rootPath, selenium: selenium});
};

exports.packagelist = function(req, res){
    res.send(_.chain(defs)
              .map(function(d){return d.split(path.sep).slice(-1)[0];})
              .filter(function(d){return d.indexOf("def") !== -1;}).value());
};

exports.rendering = function(req, res){
    if(req.params.file === "packages"){
        defs = "";
        updatedefs(
            function(){
                res.render(req.params.file, {
                    files: _.chain(defs)
                            .map(function(d){return d.split(path.sep).slice(-1)[0];})
                            .filter(function(d){return d.indexOf("def") !== -1;}).value(),
                    colors: ["Classic", "Sea", "Fire", "Dark"]
                });
            }
        );
    }
    else if(req.params.file === "version"){
        exec('batchflow -v', function(err, stdout, stderr){
            if(err !== null)
                res.send(200, "batchflow is not installed");
            else{
                if(stderr.indexOf("hark-ros.tlb") !== -1){
                    stdout += "HARKROS\n";
                }
                res.send(200, stdout.replace("\n", "<br>"));
            }
        });
    }
    else {
        res.render(req.params.file);
    }
};

exports.getFile = function(req, res){
    lock();
    var reader = csv.createCsvFileReader('networks.csv');
    reader.setColumnNames(['name', 'type']);
    reader.addListener('data', function(data){
        if(data.name == req.params.name){
            var xml = fs.readFileSync("userdata/networks/" + data.name,
                                      {encoding: "utf-8"});
            xml = xml.replace(/#!.*batchflow\n/, "");
            var parser = new xml2js.Parser();
            parser.addListener('end', function(result){
                data.object = xml2data(result);
                res.send(200, data);
                return;
            });
            parser.parseString(xml);
        }
    });
    unlock();
};

exports.dlFile = function(req, res){
    lock();
    var reader = csv.createCsvFileReader('networks.csv');
    var isSent = false;
    reader.setColumnNames(['name', 'type']);
    reader.addListener('data', function(data){
        if(data.name == req.params.name && data.type == req.params.type){
            console.log("userdata/networks/" + data.name);
            res.download("userdata/networks/" + data.name);
            isSent = true;
        }
    });

    // for files that are not in URL
    reader.addListener('end', function(data){
        if(req.params.type === "result")
            res.download("userdata/networks/" + req.params.name);
        else if(!isSent)
            res.send(404, "File not found");
    });
    unlock();
};

var isNotDir = function(filepath){return !isDir(filepath);};
var isDir = function( filepath ) {
    if( /\/$/.test( filepath ) ) {
        filepath = filepath.substr( 0, filepath.lastIndexOf('/')-1 );
    }
    return (fs.existsSync( filepath ) && fs.statSync( filepath ).isDirectory());
};
exports.getAllFiles = function(req, res){
    lock();
    var reader = csv.createCsvFileReader('networks.csv');
    reader.setColumnNames(['name', 'type']);

    var result = [];
    reader.addListener('data', function(data){
        result.push(data);
    });
    reader.addListener('end', function(){
        console.log(_.pluck(result, "name"));

        // after listing all files in .csv, get all result files.
        var existingFileNames = _.pluck(result, "name");
        var files = _.chain(fs.readdirSync("userdata/networks"))
                     .filter(isNotDir)
                     .filter(function(f){return existingFileNames.indexOf(f) === -1;})
                     .each(function(f){result.push({name: f, type:"result"});})
                     .value();
        res.send(200, result);
    });
    unlock();
};

exports.deleteFile = function(req, res){
    lock();
    var reader = csv.createCsvFileReader("networks.csv");
    reader.setColumnNames(['name', 'type']);

            // if(data.name !== req.params.name){
    var result = [];
    reader.addListener('data', function(data){
        result.push(data);
    });
    reader.addListener('end', function(){
        var writer = csv.createCsvFileWriter("networks.csv");
        _.chain(result)
         .reject(function(e){return e.name == req.params.name;})
         .each(function(d){
            writer.writeRecord([d.name, d.type]);});
         fs.unlink("userdata/networks/" + req.params.name);
    });
    res.send(200, "file deleted");
    unlock();
};

exports.addFile = function (req, res) {
    lock();
    // console.log(req.body);
    var reader, fields;
    if(req.params.type == "network"){
        // check if the value is valid
        var error = checkData(req.body, "network");
        if(error !== ""){
            res.send("504", error);
            return;
        }

        reader = csv.createCsvFileReader("networks.csv");
        reader.setColumnNames(['name', 'type']);
        fields = [];
        reader.addListener('data', function(data){
            if (data.name !== req.params.name)
                fields.push(data);
        });

        console.log(req.body);

        if(req.body.content !== undefined){
            reader.addListener('end', function(){
                var parser = new xml2js.Parser();

                parser.addListener('end', function(result){
                    var writer = csv.createCsvFileWriter('networks.csv');
                    _.each(fields, function(f){
                        writer.writeRecord([f.name, f.type]);
                    });
                    writer.writeRecord([req.params.name, req.params.type]);
                    saveNetwork(req.params.name, req.body.content);
                    res.send(200, "convert succeeded");
                });
                var line = req.body.content;
                if(line)
                    parser.parseString(line.slice(line.indexOf("\n") + 1));
            });
        }
        if(req.body.object !== undefined){
            _.each(req.body.object.nodes, function(n){
                console.log(n.id, n.pos);
            });

            reader.addListener('end', function(){
                var writer = csv.createCsvFileWriter('networks.csv');
                _.each(fields, function(f){
                    writer.writeRecord([f.name, f.type]);
                });


                // misteriously an empty list dissappears.
                _.each(req.body.object.nodes, function(n){
                    _.each(n.parameters, function(p){
                        if(p.list === undefined){
                            p.list = [];
                        }
                    });
                });

                writer.writeRecord([req.params.name, req.params.type]);
                saveNetwork(req.params.name, data2xml(req.body.object));
                res.send(200, "convert succeeded");
            });
        }
    }
    else {
        reader = csv.createCsvFileReader("networks.csv");
        reader.setColumnNames(['name', 'type']);
        fields = [];
        reader.addListener('data', function(data){
            if (data.name !== req.params.name)
                fields.push(data);
        });
        // for type === tf, audio, results
        var regex = /^data:.*;base64,(.*)$/;
        var buffer = new Buffer(req.body.content.match(regex)[1], 'base64');
        fs.writeFile("userdata/networks/" + req.params.name, buffer);

        reader.addListener('end', function(){
            var writer = csv.createCsvFileWriter('networks.csv');
            _.each(fields, function(f){
                writer.writeRecord([f.name, f.type]);
            });
            writer.writeRecord([req.params.name, req.params.type]);
            res.send(200, "convert succeeded");
        });
    }
    unlock();
};

function saveNetwork(name, content){
    fs.writeFile("userdata/networks/" + name, content);
}


function data2xml(network){
    var result = [];

    // templates for xml conersion
    var tmplSheet   = '  <Network type="<%= type %>" name="<%= name %>">';
    var tmplNode    = '    <Node name="<%= id %>" type="<%= name %>" x="<%= pos[0] %>" y="<%= pos[1] %>">';
    var tmplNodeOut = '    <NetOutput name="<%= name %>" node="<%= nodeId %>" terminal="<%= termName %>" object_type="<%- object_type %>" description="<%- description %>"/>';
    var tmplNodeIn  = '    <NetInput name="<%= name %>" node="<%= nodeId %>" terminal="<%= termName %>" object_type="<%- object_type %>" description="<%- description %>"/>';
    var tmplNodeCond= '    <NetCondition name="<%= name %>" node="<%= nodeId %>" terminal="<%= termName %>"/>';
    var tmplParam   = '      <Parameter name="<%= name %>" type="<%- type %>" value="<% if(type==="object"){ %><%- value %><% } else { %><%= value %><% } %>" description="<%- description %>"/>';
    var tmplLink    = '    <Link from="<%= fromNodeId %>" output="<%= fromTermName %>" to="<%= toNodeId %>" input="<%= toTermName %>"/>';

    // headers
    result.push('#!/usr/bin/env batchflow');
    result.push('<?xml version="1.0"?>');
    result.push("<Document>");

    // generate for each sheets
    var nodesPerSheet = _.groupBy(network.nodes, "sheet");
    _.each(_.keys(nodesPerSheet), function(sheet){
        result.push(_.template(tmplSheet, network.sheets[sheet]));

        // generate nodes and their parameters
        _.each(_.filter(network.nodes, function(node){return node.sheet == sheet;}),
            function(node){
            result.push(_.template(tmplNode, node));
            _.chain(node.parameters)
             .each(function(parameter){
                parameter.description = parameter.description.replace(/[ \n\r]*$/g, "");
                result.push(_.template(tmplParam, parameter));
             });
            result.push("    </Node>");
        });

        // generate links
        _.each(_.filter(network.links, function(link){return link.sheet == sheet;}),
            function(link){result.push(_.template(tmplLink, link));}
        );

        // generate network I/O/Condition
        _.each(_.filter(network.terminalRoles, function(role){return role.sheet == sheet;}),
            function(role){
                switch(role.role){
                    case "input"    : result.push(_.template(tmplNodeIn  , role));break;
                    case "output"   : result.push(_.template(tmplNodeOut , role));break;
                    case "condition": result.push(_.template(tmplNodeCond, role));break;
                }
            }
        );
        result.push("  </Network>");
     });

    // footers
    result.push("</Document>");
    result.push("");
    return result.join("\n");
}

function xml2data(xml){
    // console.log(util.inspect(xml, depth=10));
    var sheets = [];
    _.each(xml.Document.Network, function(network){
        // console.log(util.inspect(network, depth=10));
        var sheet = {name: network.$.name,
                     type: network.$.type,
                     links: [],
                     nodes: [],
                     roles: [],
                     notes: []};

        // nodes
        _.each(network.Node, function(node){
            // console.log(util.inspect(node));
            var tmpnode = {
                id: node.$.name,
                index: parseInt(node.$.name.split("_").slice(-1), 10),
                name: node.$.type,
                pos: [Math.floor(parseInt(node.$.x, 10) / 10) * 10,
                      Math.floor(parseInt(node.$.y, 10) / 10) * 10],
                parameters: [],
                sheet: sheet.name
            };

            _.each(node.Parameter, function(param){
                // console.log(util.inspect(param));
                tmpnode.parameters.push({
                    name: param.$.name,
                    type: param.$.type,
                    list: [],
                    value: param.$.value,
                    description: param.$.description
                });
            });
            sheet.nodes.push(tmpnode);
        });

        // links
        _.each(network.Link, function(link){
            // console.log(util.inspect(link));
            var tmplink = {
                fromNodeId: link.$.from,
                fromTermName: link.$.output,
                toNodeId: link.$.to,
                toTermName: link.$.input,
                sheet: sheet.name
            };
            sheet.links.push(tmplink);
        });

        // terminalRoles
        _.each(network.NetInput, function(term){
            console.log(term);
            sheet.roles.push({
                name       : term.$.name,
                nodeId     : term.$.node,
                description: term.$.description,
                object_type: term.$.object_type,
                role       : "input",
                sheet      : sheet.name,
                termName   : term.$.terminal
            });
        });

        _.each(network.NetOutput, function(term){
            console.log(term);
            sheet.roles.push({
                name       : term.$.name,
                nodeId     : term.$.node,
                description: term.$.description,
                object_type: term.$.object_type,
                role       : "output",
                sheet      : sheet.name,
                termName   : term.$.terminal
            });
        });

        _.each(network.NetCondition, function(term){
            console.log(term);
            sheet.roles.push({
                name       : term.$.name,
                nodeId     : term.$.node,
                role       : "condition",
                sheet      : sheet.name,
                termName   : term.$.terminal
            });
        });

        // Node
        _.each(network.Note, function(note){
            // console.log(util.inspect(note));
            var tmpnote = {
                pos: [parseInt(note.$.x, 10), parseInt(note.$.y, 10)],
                visible: parseInt(note.$.visible, 10),
                text: note.$.text
            };
            sheet.notes.push(tmpnote);
        });
        sheets.push(sheet);
    });

    // make inputs for dynamic nodes
    var dynamicNodeNames = _.chain(sheets)
                            .pluck("name")
                            .reject(function(n){return n === "MAIN";})
                            .value();
    var dynamicNodes = _.chain(sheets)
                        .pluck("nodes")
                        .map(function(nodes){
                            return _.filter(nodes, function(n){
                                return dynamicNodeNames.indexOf(n.name) !== -1; });
                        })
                        .flatten()
                        .value();
    _.each(dynamicNodes, function(dyn){
        var sheet = _.find(sheets, function(s){return s.name === dyn.name;});
        dyn.inputs = _.chain(sheet.roles)
                      .filter(function(t){return t.role === "input";})
                      .map(function(t){
                        return {
                            description: t.description,
                            name       : t.name,
                            type       : t.object_type};
                      })
                      .value();
        dyn.outputs = _.chain(sheet.roles)
                      .filter(function(t){return t.role === "output";})
                      .map(function(t){
                        return {
                            description: t.description,
                            name       : t.name,
                            type       : t.object_type};
                      })
                      .value();
    });

    // fix types
    var network = {};
    network.sheets = {};
    _.each(sheets, function(s){network.sheets[s.name] = {name: s.name, type: s.type};});
    network.nodes  = _.chain(sheets).pluck("nodes").flatten().value();
    network.links  = _.chain(sheets).pluck("links").flatten().value();
    network.terminalRoles = _.chain(sheets).pluck("roles").flatten().value();
    network.notes  = _.chain(sheets).pluck("notes").flatten().value();

    // adjust pos for each sheet
    // 1. shift
    _.each(_.keys(network.sheets), function(sheet){
        var common = _.chain(network.nodes)
                      .filter(function(n){return n.sheet === sheet;})
                      .pluck("pos").value();
        var minx = _.chain(common).map(function(p){return p[0];}).min().value();
        var miny = _.chain(common).map(function(p){return p[1];}).min().value();
        _.chain(network.nodes)
         .filter(function(n){return n.sheet === sheet;})
         .each(function(n){
            n.pos[0] -= minx;
            n.pos[1] -= miny;
         });
    });
    // 2. change scale
    var size = [1000, 1000]; // all node positions must be in this size
    _.each(network.nodes, function(n){
        if(n.pos[0] > size[0]) n.pos[0] = size[0];
        if(n.pos[1] > size[1]) n.pos[1] = size[1];
    });

    // re-index all nodes
    _.chain(network.nodes).groupBy("name").each(function(nodes, nodename){
        var ids = _.pluck(nodes, "id");
        // if duplicated ids exist
        if(nodes.length > 1 && ids.length !== _.uniq(ids).length) {
            var counter = 1;
            _.each(nodes, function(n){
                var newNodeId = "node_" + nodename + "_" + counter;
                console.log("newNodeId:", newNodeId);
                // rename all links
                _.chain(network.links)
                 .filter(function(l){return l.sheet === n.sheet;})
                 .each(function(l){
                    if(l.fromNodeId === n.id)
                        l.fromNodeId = newNodeId;
                    if(l.toNodeId === n.id)
                        l.toNodeId = newNodeId;
                 });

                // rename all terminalRoles
                _.chain(network.terminalRoles)
                 .filter(function(r){return r.sheet === n.sheet;})
                 .filter(function(r){return r.nodeId === n.id;})
                 .each(function(r){
                    r.nodeId = newNodeId;
                 });

                // rename all nodeids
                _.chain(network.nodes)
                 .filter(function(tmpn){return tmpn.sheet === n.sheet;})
                 .filter(function(tmpn){return tmpn.id === n.id;})
                 .each(function(tmpn){
                    n.index = counter;
                    n.id = newNodeId;
                 });
                counter += 1;
            });
        }
    });

    console.log(util.inspect(network));
    return network;
}

exports.templates = function(req, res){
    res.render(req.params.name);
};

exports.nodedefs = function(req, res){
    var def = _.find(defs, function(d){
        return d.split(path.sep).slice(-1).indexOf(req.params.name) !== -1;});

    if(def){
        // send processed data
        fs.readFile(def, function(err, data){
            if(err){
                res.send(500);
                return;
            }
            var parser = new xml2js.Parser();
            parser.addListener('end', function(result){
                var defobj = _.map(result.Definitions.NodeClass, def2node);
                // console.log(defs)

                // read .n files in .def files (#1214)
                var dirname = def.replace(def.split(path.sep).slice(-1)[0], "");
                var nfiles = _.filter(fs.readdirSync(dirname),
                                      function(f){return f.slice(-2) === ".n";});
                // console.log("readDirSync:", dirname, nfiles);

                if(nfiles.length > 0){
                    var send = _.after(nfiles.length,
                        function(){res.send(200, JSON.stringify(defobj));});
                    _.each(nfiles, function(nfile){
                        var string = fs.readFileSync(path.join(dirname, nfile), 'utf8');
                        string = string.slice(string.indexOf("\n"));
                        var tmpparser = new xml2js.Parser();
                        tmpparser.addListener('end', function(result){
                            defobj.push(nfile2node(result.Document, nfile.split(".")[0]));
                            defobj = _.sortBy(defobj, "name");
                            send();
                        });
                        tmpparser.parseString(string);
                    });
                }
                else{
                    defobj = _.sortBy(defobj, "name");
                    res.send(200, JSON.stringify(defobj));
                }
            });
            parser.parseString(data);
        });
    }
    else{
        console.log("def not found");
        res.send(404);
    }
};
function nfile2node(nfile, filename){
    var node = {
        category: nfile.$.category,
        description: "",
        inputs: [],
        outputs:[],
        name: filename,
        parameters: []
    };

    node.inputs = _.map(nfile.Network[0].NetInput, function(inp){
        return {
            description: "",
            name: inp.$.name,
            object_type: "any"
        };
    });

    node.outputs = _.map(nfile.Network[0].NetOutput, function(out){
        return {
            description: "",
            name: out.$.name,
            object_type: "any"
        };
    });

    node.parameters = _.map(nfile.Parameter, function(tmp){
        return {
            description: "",
            condition: "",
            list: [],
            name: tmp.$.name,
            type: (tmp.$.type === "")? "string" : tmp.$.type,
            value: tmp.$.value
        };
    });
    return node;
}

function def2node(def){
    var node = {
        category: def.$.category,
        description: def.Description[0],
        inputs: [],
        outputs:[],
        name: def.$.name,
        parameters: []
    };
    node.inputs = _.map(def.Input, function(tmp){
        return { description:tmp._,
                 name:tmp.$.name,
                  object_type: tmp.$.type
                };
    }) || [];

    node.outputs = _.map(def.Output, function(tmp){
        return { description:tmp._,
                 name:tmp.$.name,
                  object_type: tmp.$.type
                };
    }) || [];

    node.parameters = _.map(def.Parameter, function(tmp){
        return {
            description: tmp._,
            condition: tmp.$.valid,
            list: (tmp.$.list === "")? [] : tmp.$.list.split(":"),
            name: tmp.$.name,
            type: (tmp.$.type === "")? "string" : tmp.$.type,
            value: tmp.$.value
        };
    });

    return node;
}


var child;
var child_result = "";
var stderr = "";
var stdout = "";
var whole = "";
exports.execution = function(req, res){
    if(child !== undefined){
        console.log("child is not undefined");
        res.send(504, "Another process already exists");
        return;
    }

    var cmd = 'batchflow ' + req.params.file;
    console.log(cmd);

    stderr = "";
    stdout = "";
    whole = "";

    child = spawn("batchflow", [req.params.file],
        {env: process.env, cwd: "userdata/networks", killSignal: "SIGTERM"});

    child.stderr.on("data", function(data){
        stderr += data + "\n"; whole += "<stderr>" + data + "</stderr>\n";
    });
    child.stdout.on("data", function(data){
        stdout += data + "\n"; whole += "<stdout>" + data + "</stdout>\n";
    });
    child.on("close", function(code, signal){
        if(code === 0)
            child_result = "Normally finished.";
        else
            child_result = "Stopped";
        child = undefined;
    });

    // child.unref();
    res.send(200, "Process started");
};

exports.execstatus = function(req, res){
    if(child !== undefined)
        res.send(200, "Process is running.");
    else
        res.send(200, child_result + "\n" + whole);
};

exports.abort = function(req, res){
    if(child){
        console.log(child.pid);
        child.kill('SIGKILL');
        res.send(200, "Killed " + whole);
        child = undefined;
    }
    else{
        res.send(404, "No process is running");
    }
};

exports.os = function(req, res){
   res.send(200, distribution);
};

exports.keepalive = function(req, res){
    res.send(200);
};


exports.reshape = function(req, res){
    if(!dotPath){
        console.log(dotPath);
        res.send(404, "dot not found. set PATH variable");
        return;
    }

    var network = JSON.parse(JSON.stringify(req.body));
    var filename = "dot/test.dot";

    var ret;
    if(distribution === "Linux" || distribution === "Darwin")   ret = "\n";
    if(distribution === "Windows") ret = "\r\n";

    //convert from network to dot language
    fs.writeFileSync(filename, "");
    fs.appendFileSync(filename, "digraph network {" + ret);
    fs.appendFileSync(filename, "  graph [rankdir = LR]" + ret);
    _.each(network.sheets, function(sheet){
        // filename = "dot/" + sheet.name + ".dot";
        fs.appendFileSync(filename, "  subgraph cluster" + sheet.name + "{" + ret);
        fs.appendFileSync(filename, "    label=\"" + sheet.name + "\";" + ret);
        _.chain(network.nodes).filter(function(n){return n.sheet === sheet.name;})
         .each(function(n){
            fs.appendFileSync(filename, "    " + n.id + "[" + ret);
            fs.appendFileSync(filename, "      shape = Mrecord," + ret);
            fs.appendFileSync(filename, "      label = \"" + n.id + "|");
            fs.appendFileSync(filename, "{{");
            _.each(n.inputs, function(io){
                fs.appendFileSync(filename, util.format("<%s> %s", io.name , io.name));
                if(io !== n.inputs.slice(-1)[0])
                    fs.appendFileSync(filename, "|");
            });
            fs.appendFileSync(filename, "}|{");
            _.each(n.outputs, function(io){
                fs.appendFileSync(filename, util.format("<%s> %s", io.name, io.name));
                if(io !== n.outputs.slice(-1)[0])
                    fs.appendFileSync(filename, "|");
            });
            fs.appendFileSync(filename, "}}");
            fs.appendFileSync(filename, "\"" + ret + "    ]" + ret);
         });

        _.chain(network.links).filter(function(l){return l.sheet === sheet.name;})
         .each(function(l){
            fs.appendFileSync(filename,
                util.format("    %s:%s -> %s:%s;%s",
                            l.fromNodeId, l.fromTermName,
                            l.toNodeId, l.toTermName, ret));
         });
        fs.appendFileSync(filename, "  }\n");
    });
    fs.appendFileSync(filename, "}\n");

    // calculate positions and send it.
    var re = new RegExp(/([a-zA-Z0-9_]+)\W+\[.*pos="(\d+),(\d+)".*\]/);
    exec(dotPath + " dot/test.dot", function(error, stdout, stderr){
        if(error){
            console.log(stderr, error);
            res.send(504, stderr);
        }
        // copy the positions calculated by graphviz
        _.each(stdout.replace(/\n/g, "").replace(/\r/g, "").split(";"), function(line){
            if(line.match("node_") && line.indexOf("->") === -1){
                console.log(line);
                var result = re.exec(line);
                var nodeid = result[1];
                var left = parseInt(result[2], 10);
                var top  = parseInt(result[3], 10);
                var nodeobj = _.find(req.body.nodes, function(n){return n.id === nodeid;});
                nodeobj.pos = [left, top];
            }
        });

        // shift
        _.each(_.keys(req.body.sheets), function(sheet){
            var minx = _.chain(network.nodes)
                        .filter(function(n){return n.sheet === sheet;})
                        .map(function(n){return n.pos[0];})
                        .min().value();
            var miny = _.chain(network.nodes)
                        .filter(function(n){return n.sheet === sheet;})
                        .map(function(n){return n.pos[1];})
                        .min().value();

            _.chain(network.nodes).filter(function(n){return n.sheet === sheet;})
             .each(function(n){
                n.pos[0] = parseInt((n.pos[0] - minx) * 1.0, 10);
                n.pos[1] = parseInt((n.pos[1] - miny) * 1.0, 10);
             });
        });

        // for debug
        _.each(network.nodes, function(n){
            console.log(n.id, n.pos[0], n.pos[1]);
        });
        console.log(network);

        // restore undefined properties if not exist
        network.links = network.links || [];
        network.terminalRoles = network.terminalRoles || [];
        _.each(network.nodes, function(n){
            _.each(n.parameters, function(p){
                p.list = p.list || [];
            });
        });

        res.send(200, network);
    });
};

function checkData(obj, type){
    if(type === "network"){
        if(obj.content){
            if(typeof(obj.content) !== 'string') return "This is not a text file";
            if(obj.content.indexOf("#!") !== 0) return "This does not starts with #!";
            if(obj.content.indexOf("<?xml") === -1) return "This not an xml file";
            if(obj.content.indexOf("batchflow") === -1) return "No word 'bathcflow' found";
            return "";
        }
        else{
            return "";
        }
    }
    else{
        return "unknown data type";
    }
    return "This should not happen";
}

exports.datapath = function(req, res){
    var reg = new RegExp(path.delimiter, 'g');
    var url = "file://" + process.cwd().replace(reg, "/") + "/userdata/networks";
    res.send(200, url);
};

exports.login = function(req, res){
    var css = ['jquery/jquery-ui-1.10.2.custom/css/smoothness/jquery-ui-1.10.2.custom.min.css'];
    var js =  [
               'js/json2.js',
               'jquery/jquery.min.1.9.1.js',
               'jquery/jquery-ui-1.10.2.custom/js/jquery-ui-1.10.2.custom.min.js',
               'jquery/jquery.ui.position.js',
               'js/underscore.js',
               'js/arrayObjectMixin.js',
               'js/backbone.js',
               'js/fileBrowse.js',
               'js/login.js'];

    css = _.map(css, function(tmp){return rootPath + tmp;});
    js  = _.map(js , function(tmp){return rootPath + tmp;});
    res.render('login', {css: css, javascripts: js});
};

exports.checklogin = function(req, res){
    email = req.body.email;
    password = req.body.password;

    console.log(req.body);
    if(req.params.type === "signup"){
        //signup
        var reader = csv.createCsvFileReader("users.csv");
        reader.setColumnNames(['email', 'password']);
        reader.addListener('data', function(data){
            if(data.email === email){
                res.send(404);
            }
        });
        reader.addListener('end', function(){
            var writer = csv.createCsvFileWriter("users.csv");
            writer.writeRecord([email, password]);
            res.send(200);
        });
    }
    else{
        //login
        var reader = csv.createCsvFileReader("users.csv");
        reader.setColumnNames(['email', 'password']);
        reader.addListener('data', function(data){
            if(data.email === email && data.password === password){
                res.send(200);
            }
        });
        reader.addListener('end', function(){
            res.send(404);
        });
    }
}

function lock(){
    var lock = path.join(process.cwd(), "lock");
    if(!fs.exists(lock)){
        fs.writeFileSync(lock, "");
        return;
    }

    while(fs.exists(lock)){
        console.log("waiting");
    }
}

function unlock(){
    var lock = path.join(process.cwd(), "lock");
    fs.unlinkSync(lock);
}

